# discrete2unary.awk
#
# Thu Feb 21 10:42:22  2002
#
# converts a discrete data file to a unary representation
# assumes data class in last column, and (gnu)sort in path
#
# -vIGNORE='RANGE'
# where RANGE is 
#  0. default "": none ignored
#  1. an index indicating the column to be ignored, 0 for the last column
#  2. a comma-separated list of indices
#  3. a range in a-b notation, meaning all indices between a and b including a,b
#  4. a comma-separated list of ranges
#  5. a comma-separated list of ranges and indices
#
# This script will use a large amount of memory, if the source is large.
#
# Example:
#   zcat mushroom_data.txt.gz | awk -F, -f discrete2unary.awk - > mushroom_unary.dat 2> mushroom_utrans.dat
# 
#   converts the all of the ,-separated mushroom.txt columns to unary coding
#   to mushroom_unary.dat
#
#   prints block lengths, number of data samples, dimension (without class),
#   and a transscript symbol table of dimension, block number, decimal code,
#   symbol from source, and unary code to the standard error channel.

BEGIN {

  TOSTDERR = "cat >&2"

  if(IGNORE != "") {
    
    split(IGNORE, rnges, ",")
    
    for(rng in rnges) {
      
      no = split(rnges[rng], fromto, "-")

      if(no == 1) 
        from = to = fromto[1] + 0.
      else if(no == 2)
        { from = fromto[1] + 0.; to = fromto[2] + 0.}
      else
        { printf "Error in IGNORE arg!" | STDERR; exit(0) }
      
      if(from > to) { h = from; from = to; to = h }

      for(i = from; i <= to; i++) ++ignore[i]
      
    }
  }
}

!/(^#)|(^[ \t]*$)/ {

  ++nr
  nf = NF

  for(i = 1; i <= nf; i++) {
  
    if(!(i in ignore || (i==nf && 0 in ignore))) {
      m = ++cnt[i,$i]
      if(m == 1) symb[i,$i] = ++res[i]
      data[nr,i] = symb[i,$i]
    }
    else {
      res[i] = 1
      data[nr,i] = $i
    }
  }
}

END {

  printf(">") | TOSTDERR
  for(i = 1; i < nf; i++) { 
    printf(" %d", res[i]) | TOSTDERR
    sum += res[i] 
  }
  printf(" : #%d %d\n", nr, sum) | TOSTDERR
  
  for(s in symb) {
    
    split(s,ss,SUBSEP)
    
    if(ss[1] != nf)
      print ss[1], symb[s], ss[2], unary(symb[s],res[ss[1]])  | "sort -t' ' -r -k2 -n | sort -t' ' -s -k1 -n | cat -n >& 2"
  }

  for(n = 1; n <= nr; n++) {
  
    for(i = 1; i < nf; i++)
      printf("%s ", (i in ignore) ? data[n,i] : unary(data[n,i],res[i]))

    printf(" %s\n", ((0 in ignore) ? data[n,nf] : data[n,nf] - 1))
  }
}

function unary(k,max) {

  s = ""

  for(t = 1; t <= max; t++) s = ((t == k) ? "1 " : "0 ") s

  return s
}
