/*
 #   Supervised Relevance Neural Gas (srng) package.
 #   Copyright (C) 2002  Marc Strickert. (mstricke@uos.de)
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
 #   the Free Software Foundation, version 2 of the License.
 #
 #   This program is distributed in the hope that it will be useful,
 #   but WITHOUT ANY WARRANTY; without even the implied warranty of
 #   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #   GNU General Public License (LICENSE.txt) for more details.
 #
 #   Marc Strickert, 25 February 2002
 */



/*
 # gcc -O4 -Wall -o srng srng.c -lm
 #
 # srng.c, Fri Jan 11 09:14:50  2002
 # 
 # the source is not quite ANSI C:
 #   warning: ANSI does not permit the keyword `inline'
 #   warning: ANSI C forbids variable-size array `xxx'
 #
 # Tue Jan 15 09:35:51  2002: 
 #     - Barbara's help(rank correct scaling) -> rnk_cs = 1./help
 #     - Denominator threshold in srng.h set to 1e-13, Warning treatment
 #
 # Wed Jan 16 12:09:20  2002
 #     - some speedups (1-,5-neig, qsort), rank-based adaptation
 #
 # Fri Jan 18 07:46:09  2002
 #     - Discrete mode: lambda-blocks
 #     - new parameter in init file added: 
 #         neighborhood_size - <0. -> prototypes_no / 0. -> NEIGEPS
 #     - new parameter in file file added: 
 #         randseed - 0 -> randomize() 
 #     
 #     prototypes_no, 
 #     rate_eta, rate_alpha, rate_alphadecay, 
 #     rate_neighborhood_annealing, neighborhood_size, ng_fader, 
 #     randseed
 #
 # Wed Jan 30 10:24:20  2002
 #   - command argument 1: init_filename 
 #        filename may be -, meaning stdin (after reading pattterns from stdin)
 #        
 #      mode 1: file contains one single initialization record, 
 #               -1 lambda and no blocklens allowed
 #      mode 2: file contains continuous records,
 #               full lambda and blocklen specification required
 #      
 #      Example to read from work-in-progress mushroomZ_cb.dat:
 #       > tail --lines=+1 -f mushroomZ_cb.dat | \
 #           cat mushroomZ.dat - | \
 #            ./srng - 0 | grep '^-1 '
 #      
 #   - command argument 2: no_cycles
 #      if  > 0: train
 #         == 0: error with original prototype output
 #          < 0: error with prototype output, only if pt_class_specificity > 50%
 #
 #   - command argument 3: moduloprint
 #      if argument 2:no_cycles <= 0 -> select record #moduloprint
 */



/******************************************************************************/


#include "srng.h"


inline double sgd(double x) {   /* speedup: linearly sampled sigmoid function */

  static double y[] = 
    { 4.53979e-005, 5.55606e-005, 6.79983e-005, 8.322e-005,
      0.000101849, 0.000124647, 0.000152548, 0.000186693,
      0.000228479, 0.000279615, 0.000342191, 0.000418767,
      0.000512469, 0.000627125, 0.000767413, 0.000939055,
      0.00114904, 0.00140592, 0.00172013, 0.0021044,
      0.00257431, 0.00314881, 0.00385103, 0.00470911,
      0.00575729, 0.00703712, 0.00859899, 0.0105038,
      0.0128252, 0.0156515, 0.0190885, 0.0232625,
      0.0283229, 0.0344452, 0.0418339, 0.0507244,
      0.0613831, 0.0741067, 0.0892171, 0.107052,
      0.127952, 0.152236, 0.180177, 0.211963,
      0.247664, 0.287186, 0.330246, 0.376355,
      0.424817, 0.474769 };
  
  static int i;
  static double d, y1, y2;
  
  x -= 0.1; /* due to integer rounding */

  if(x < -10.) return 0.;
  if(x >= 10.) return 1.;

  i = (int)(d = (x + 10.) * 5.);

  d -= i;
  

  y1 = (i >= 50) ? (1. - y[99-i]) : y[i];

  if(i == 99) 
    y2 = 1.;
  else
    y2 = (i >= 49) ? (1. - y[98-i]) : y[i+1];

  return ((1.-d) * y1 + d * y2);
}


inline double sgd_prime(double x)
{
  static double t;
  t = sgd(x);          /* t=1/(1.+exp(-x)); */ /* for testing purposes. OK! */
  return t * (1. - t);
}


/******************************************************************************/



void prototype_free(PROTOTYPE n) 
{
  free(n -> coord);
  free(n -> coord_shaddow);
  if(n -> weights != global_lambda) free(n -> weights);
  free(n);
}



void ptfree() 
{
  while(--prototypes_lastindex >= 0) 
    prototype_free(prototypes[prototypes_lastindex]);
  free(prototypes);
  free(global_lambda);
  free(block_len);
  free(block_weight);
  free(prototypes_no_class);

  prototypes = NULL, global_lambda = NULL,
    prototypes_no_class = NULL, block_len = NULL, block_weight = NULL;
  
  prototypes_lastindex = 0;
}



PROTOTYPE prototype_new_empty() 
{
  
  PROTOTYPE n = (PROTOTYPE)malloc(sizeof(PROTOTYPE_TYPE));

  if(n == NULL) panic("prototype_new_empty(): cannot malloc prototype!");
  
  if(prototypes_lastindex == 0 &&
      (prototypes = 
         (PROTOTYPE *)malloc(sizeof(PROTOTYPE) * prototypes_no)) == NULL)
    
      panic("prototype_new_empty(): cannot malloc prototypes!");


  prototypes[prototypes_lastindex++] = n; /* XXX: very courageous here */

  return n;
}



PROTOTYPE prototype_new(int cls)
{
  
  PROTOTYPE n = prototype_new_empty();
  
  if((n -> coord = (double *)malloc(sizeof(double) * (pattern_dimension+1))) == NULL)
    panic("prototype_new(): cannot malloc coord!");
  
  if((n -> coord_shaddow = (double *)malloc(sizeof(double) * (pattern_dimension+1))) == NULL)
    panic("prototype_new(): cannot malloc coord_shaddow!");

  n -> cls = cls;
  n -> weights = global_lambda; 
  n -> dist = 0.;

  return n; 
}



inline double prototype_patterndistance(PROTOTYPE n, double *pattern)
{

  int i = pattern_dimension - 1;

  double *w = n -> coord,
         *lmb = n -> weights,
         res = 0., tmp;
  
  for(; i >= 0; i--) tmp = w[i] - pattern[i], res += lmb[i] * tmp * tmp;
  
  return res;
}



inline int prototype_dists_cmp(void const *a1, void const *a2) { 

  static double d1, d2;
  
  if((d1 = (*(PROTOTYPE *)a1) -> dist) < (d2 = (*(PROTOTYPE *)a2) -> dist)) 
    return -1;
  else if(d1 > d2) 
    return 1;

  return 0;
}



/*
 * take all prototypes, calc distances to
 * presented pattern, sort correct and rank.
 * With speedups (mode-dependent).
 * Thereby, watch out for nearest wrong prototype.
 * return results in **correct (possibly fresh allocated), and wrong, resp.
 *
 * return number of correct prototypes
 */

int prototype_prepare_by_pattern(
    double *pattern, PROTOTYPE **correct, PROTOTYPE *wrong, int mode) 
{

  double min_wrong = DBL_MAX, min_correct = DBL_MAX, d; 

  int i,
      cls = (int)pattern[pattern_dimension],
      no_correct = 0;

  PROTOTYPE pt, *result; 


  if(*wrong == NULL && 
     (*correct=(PROTOTYPE *)malloc(sizeof(PROTOTYPE) * prototypes_no)) == NULL) 
    panic("prototype_prepare_by_pattern(): cannot malloc() result");
  

  result = *correct;


  for(i = 0; i < prototypes_no;  i++) {
    
    pt = prototypes[i]; 
    
    pt -> dist = d = prototype_patterndistance(pt, pattern);


    if(pt -> cls == cls) {
      
      if(no_correct == 0) {
        min_correct = d, result[0] = pt;
        ++no_correct; continue;
      }
      
      if(mode == 0) {
        min_correct = d, result[0] = pt;
        continue;
      }
      
      
      if(mode == 1) {

        if(no_correct > 4 && d >= result[4] -> dist) continue;

        if(no_correct == 1) {
          if(d < min_correct) {
            result[1] = result[0];
            min_correct = d, result[0] = pt;
          }
          else
            result[1] = pt;
          ++no_correct; continue;
        }
        
        if(no_correct == 2) {
          
          if(d < result[1] -> dist) {
            
            result[2] = result[1];
            
            if(d < min_correct) 
              result[1] = result[0], result[0] = pt, min_correct = d;
            else
              result[1] = pt;
          }
          else
            result[2] = pt;

          ++no_correct; continue;
        }

        if(no_correct == 3) {
          
          if(d < result[1] -> dist) {
            
            result[3] = result[2], result[2] = result[1];

            if(d > min_correct) 
              result[1] = pt;
            else
              result[1] = result[0], result[0] = pt, min_correct = d;
          }
          else {
            if(d < result[2] -> dist) 
              result[3] = result[2], result[2] = pt;
            else
              result[3] = pt;
          }

          ++no_correct; continue;
        }
          
        if(no_correct >= 4) { /* binary insert */
          
          if(d > result[2] -> dist) {
            
            if(d > result[3] -> dist) 
              result[4] = pt;
            else
              result[4] = result[3], result[3] = pt;
          }
          else { /* d <= result[2] -> dist */
            
            result[4] = result[3], result[3] = result[2]; 
            
            if(d > result[1] -> dist) 
              result[2] = pt;
            else if(d > min_correct) 
              result[2] = result[1], result[1] = pt;
            else
              result[2] = result[1], result[1] = result[0], result[0] = pt,
              min_correct = d;
          }

          no_correct = 5; continue;
        }
      }
      
      /* mode >= 2 */
      result[no_correct++] = pt;
      
    } else {     

      if(d < min_wrong) {
        min_wrong = d;
        *wrong = pt;
      }
    }
  }
  
  if(mode >= 2)                 /* sort result's correct prototypes */
    qsort(result, no_correct, sizeof(PROTOTYPE), prototype_dists_cmp);
 
  return no_correct;
}



double rank(int i) 
{

  static double ns_buf = -1.;
  static double *tab;  

  double y = 1.;

  int pmax = (prototypes_maxperclass < 5) ? 5 : prototypes_maxperclass;
  
  
  if(ns_buf < 0.  && (tab = malloc(sizeof(double) * pmax)) == NULL)
    panic("rank(): cannot malloc tab!");
  
  
  if(i < 0) { ns_buf = -1; free(tab); return 0.; }


  if(neighborhood_size != ns_buf) {
    int i;
    for(i = 0; i < pmax; i++) tab[i] = -1.;
    ns_buf = neighborhood_size;
  }
  
  
  if(tab[i] >= 0.) return tab[i];
  
  if(i > 0) {

    if((y = i/neighborhood_size) > 10.)
      y = 1e-5;
    else
      y = exp(-y);
  }

  return tab[i] = y; 
}



void prototype_update(
    double *pattern, int no_correct,
    PROTOTYPE *correct, PROTOTYPE wrong)
{
 
  int i, i_c, no_denominator_problems;

  double d,
         sum,
         rnk_c, 
         rnk_cs, 
         h1, h2,
         d_c, d_w, 
         d_cc, d_cw, 
         d_cpw, d_cmw,
         rnk[no_correct],
         *coord_c, *coord_w,
         *coord_cs, *coord_ws,
         local_lambda[pattern_dimension];

         
  no_denominator_problems = 0;


  /* copy lambda weights and coordinates */
  for(i = 0; i < pattern_dimension; i++) 
    local_lambda[i] = global_lambda[i];
  
  for(i_c = 0; i_c < prototypes_no; i_c++) 
    for(i = 0; i < pattern_dimension; i++) 
      prototypes[i_c] -> coord_shaddow[i] = prototypes[i_c] -> coord[i];
    
    
  sum = rnk_cs = 0.;

  for(i_c = 0; i_c < no_correct; i_c++) 
    if((rnk_c = rnk[i_c] = rank(i_c)) < NEIGSPEEDTHRESH) {
      no_correct = i_c; 
      break;
    }
    else
      rnk_cs += rnk_c;
  
  rnk_cs = 1. / rnk_cs;

  
  coord_w = wrong -> coord;
  coord_ws = wrong -> coord_shaddow;

  d_w = wrong -> dist;


  for(i_c = 0; i_c < no_correct; i_c++) {

    rnk_c = rnk[i_c] * rnk_cs;
    
    coord_c = correct[i_c] -> coord;
    coord_cs = correct[i_c] -> coord_shaddow;

    d_c = correct[i_c] -> dist;
    
    d_cpw = d_c + d_w;
         
    if(d_cpw > EPS) {
      
      d_cmw = d_c - d_w;
    

      h1 = h2 = sgd_prime(d_cmw / d_cpw) / (d_cpw * d_cpw);

      
      h1 *= d_w * rnk_c;  
      h2 *= d_c * rnk_c;


      for(i = 0; i < pattern_dimension; i++) {
        
        if((d = global_lambda[i]) == 0.) continue; 

        d_cc = pattern[i] - coord_c[i];
        d_cw = pattern[i] - coord_w[i];

        d_cmw = h1 * d_cc;
        d_cpw = h2 * d_cw;

        if((d *= rate_eta) > 0.) {
          coord_cs[i] += d * d_cmw;
          coord_ws[i] -= d * d_cpw * ng_fader;
        }
        
        /********* Adaptation of lambdas **********************/
        
        if((d = block_weight[i] * rate_alpha) > 0.)  {
        
          d_cmw = d * (d_cmw * d_cc - d_cpw * d_cw);

          if((local_lambda[i] -= d_cmw) < 0.) {
            d_cmw -= local_lambda[i];
            local_lambda[i] = 0.;
          }

          sum += (d_cmw < 0.) ? -d_cmw : d_cmw;
          
        }
      }
    }
    else 
      ++no_denominator_problems;
  }

  if(no_denominator_problems > 0) 
    fprintf(stderr, "Warning: %d of %d small denominator!\n", 
      no_denominator_problems, no_correct);
  
  
  if(sum != 0.) {  /* sum == 0 only, if eta == 0. */

    int k;
    
    sum = 0.;

    for(i = 0; i < pattern_dimension; i++) 
      sum += 
        (local_lambda[i] += rate_alphadecay * local_lambda[i]*local_lambda[i]);

    if(sum != 0.)  sum = 1. / sum;

    for(i = k = i_c = 0; i < pattern_dimension; i++) 
      global_lambda[i] = local_lambda[i] * sum;
  }

  
  for(i_c = 0; i_c < prototypes_no; i_c++) 
    for(i = 0; i < pattern_dimension; i++) 
      prototypes[i_c] -> coord[i] = prototypes[i_c] -> coord_shaddow[i];
}



void prototype_stdout(PROTOTYPE n)
{
  
  int i;
  
  for(i = 0; i < pattern_dimension; i++) 
    printf("%g ", n -> coord[i]);

  printf("%d\n", n -> cls);
}



/******************************************************************************/



void pattern_free(int i) 
{
  while(--i >= 0) free(pattern[i]);
  free(pattern);
}



int pattern_alloc(void) 
{

  int i, tdim = sizeof(double) * (pattern_dimension+1);
  
  if(NULL == (pattern = (double **)malloc(sizeof(double *) * pattern_length)))
    panic("pattern_alloc(): cannot malloc pattern!");

  for(i = 0; i < pattern_length; i++) {
    if(NULL == (pattern[i] = (double *)malloc(tdim))) {
      pattern_free(i);
      panic("pattern_alloc(): cannot malloc pattern[k]!");
    }
  }
  
  return 1;
}



void pattern_stdin(void) 
{
  
  int line, column;
  
  
  if((line=fgetc(stdin)) != '#') ungetc(line, stdin);  /* file starts with # */
  
  if(2 != scanf("%d %d", &pattern_length, &pattern_dimension)) 
    error("main(): Could not read proper table dimension !");
  
  if(!pattern_alloc())
    error("main(): pattern_alloc-Error!");
  
  for(line = 0; line < pattern_length; line++) {
    for(column = 0; column <= pattern_dimension; column++) {
      if(1 != scanf("%lf", pattern[line]+column)) {
        panic("main(): Trouble reading data!");
      }
    }
  }
}



/******************************************************************************/
/*              Some data shuffling routines                                  */
/******************************************************************************/


void shuffle_index_free(void) 
{
  free(shuffle_index);
}



void shuffle_index_alloc(void) 
{

  int len = pattern_length;

  shuffle_index_free();
  
  if((shuffle_index = (int *)malloc(sizeof(int) * len)) == NULL) 
    panic("shuffle_index_alloc(): cannot malloc() shuffle_index");
  
  while(--len >= 0) shuffle_index[len] = len;
}



void shuffle_do() 
{

  int cnt, ind, tmp;
  
  if(shuffle_index == NULL) 
    shuffle_index_alloc();
  
  for(cnt = 0; cnt < pattern_length; cnt++) {
    ind = irand(pattern_length);
    tmp = shuffle_index[cnt];
    shuffle_index[cnt] = shuffle_index[ind];
    shuffle_index[ind] = tmp;
  }
}



int shuffle_next(void) 
{

  static int ind = INT_MAX-1;

  if(++ind >= pattern_length) {
    shuffle_do();
    ind = 0;
  }

  return shuffle_index[ind];
}


/******************************************************************************/



void srng_print(void) 
{
  int i;

  printf("%d %g %g %g %g %g %g %u\n", prototypes_no, rate_eta, rate_alpha, rate_alphadecay, rate_neighborhood_annealing, neighborhood_size, ng_fader, randseed);

  for(i = 0; i < prototypes_no; i++) 
    prototype_stdout(prototypes[i]);

  for(i = 0; i < pattern_dimension; i++) 
    printf("%g ", global_lambda[i]);

  printf("\n");
  
  for(i = pattern_dimension; i < (pattern_dimension<<1); i++) 
    if(block_weight[i] >= 0.)
      printf("%g ", block_weight[i]);
    else
      break;
  
  printf("\n\n\n"); fflush(stdout);
}
 


void srng_init(FILE *fp) 
{

  PROTOTYPE pt;
  int i, p, readfail, t, u;
  double sum_lambdaweight, sum_lambda;
  
  
  if(8 != fscanf(fp, "%d %lf %lf %lf %lf %lf %lf%u", &prototypes_no, &rate_eta, &rate_alpha, &rate_alphadecay, &rate_neighborhood_annealing, &neighborhood_size, &ng_fader, &randseed))
    panic("srng_init(): cannot read first 8 numbers from initialization file");

  
  if((prototypes_no_class = (int *)malloc(sizeof(int) * prototypes_no)) == NULL)
    panic("srng_init(): cannot malloc() prototypes_no_class");
  
  for(p = 0; p < prototypes_no; p++) prototypes_no_class[p] = 0;

  
  if((block_len = (int *)malloc(sizeof(int) * (pattern_dimension + 1))) == NULL)
    panic("srng_init(): cannot malloc() block_len");
  
  if((block_weight = (double *)malloc(2 * sizeof(double) * (pattern_dimension + 1))) == NULL)
    panic("srng_init(): cannot malloc() block_weight");
  
  
  /* global_lambda must be known before prototype allocation */
  if((global_lambda = (double *)malloc(sizeof(double) * pattern_dimension)) == NULL) 
    panic("srng_init(): cannot malloc() global_lambda");

  
  for(p = 0; p < prototypes_no; p++) {
    
    pt = prototype_new(-1);
    
    for(i = 0; i <= pattern_dimension; i++) {
      
      if(1 != fscanf(fp, "%lf ", &(pt -> coord[i])))
        panic("srng_init(): cannot read prototype coordinates from file");
      
      pt -> coord_shaddow[i] = pt -> coord[i];
    }

    ++prototypes_no_class[ pt -> cls = (int)pt -> coord[pattern_dimension] ];
  }

  
  for(i = 0; i < pattern_dimension; i++) 
    if(1 != fscanf(fp, "%lf ", global_lambda+i))
      panic("srng_init(): cannot read lambdas from file");
    else if(global_lambda[0] < 0.)
      break;                          /* Auto-Init */
    else if(i > 0 && (global_lambda[i] < 0. || global_lambda[i] > 1.)) 
      panic("srng_init(): lambda < 0 or lambda > 1 error");

  
  /* try to read blocklens from file */
  blocks_no = readfail = 0;

  for(p = i = 0; p < pattern_dimension; i++) {
    
    if(1 != fscanf(fp, "%lf ", block_weight+i)) {
      
      if(i == 0) 
        ++readfail;
      else if(readfail == 0 && p != pattern_dimension)
        panic("srng_init(): missing lambda weights!");

      block_len[i] = readfail;
      block_weight[i] = 1.;

      if(readfail) ++blocks_no;

    }
    else {

      ++blocks_no;
      block_len[i] = (block_weight[i] > 1.) ? (int)block_weight[i] : 1;

      if(block_weight[i] < 0.)
        panic("srng_init(): lambda weights must be >= 0.!");
      
      if(block_weight[i] > 1 && block_weight[i] != block_len[i])
        panic("srng_init(): lambda weights > 1 must be integer blocklens!");

    }
    block_weight[i+pattern_dimension] = block_weight[i];

    p += block_len[i];

  }

  
  block_weight[i+pattern_dimension] = -1.;
  
  sum_lambda = sum_lambdaweight = 0.;

  readfail = global_lambda[0] < 0.;
  

  for(i = p = t = 0; i < pattern_dimension; i++) {
    
    if(i == t) {
      
      if(block_weight[i] > 1.) {

        block_weight[i] = 1. / block_weight[i];
        for(u = pattern_dimension - block_len[p]; u > i; u--)
          block_weight[u+block_len[p]-1] = block_weight[u];
          
      }
      
      t += block_len[p++];
    }
    else
      block_weight[i] = block_weight[i-1];
    
    if(readfail) global_lambda[i] = block_weight[i] / pattern_dimension;

    sum_lambdaweight += block_weight[i];
    sum_lambda += global_lambda[i];
  }

  
  if(sum_lambda == 0.)        panic("srng_init(): all lambdas zero!");
  if(sum_lambdaweight == 0.)  panic("srng_init(): all lambda weights zero!");
  
  sum_lambda = 1. / sum_lambda;
  sum_lambdaweight = 1. / sum_lambdaweight;
  
  for(i = 0; i < pattern_dimension; i++) {
    block_weight[i] *= sum_lambdaweight;
    global_lambda[i] *= sum_lambda;
  }
  
  
  prototypes_maxperclass = 0;
  for(p = 0; p < prototypes_no; p++) 
    if(prototypes_maxperclass < prototypes_no_class[p]) 
      prototypes_maxperclass = prototypes_no_class[p];
  
  
  training_cycles = 0;
  
  if(neighborhood_size < 0.)
    neighborhood_size = (double)prototypes_no;
  else if(neighborhood_size == 0.)
    neighborhood_size = NEIGEPS;
    
  if(randseed != 0)
    srand(randseed);
  else
    randseed = randomize();
    
}

  

void srng_train(int cycles)
{
  
  PROTOTYPE *corrects, wrong = NULL;

  double *pat, r = 0.;

  int p, tmp, ptnopc,
      rind = (prototypes_no > 4) ? 4 : prototypes_no;


  for(; cycles > 0; cycles--, ++training_cycles) {
  
    if(neighborhood_size > NEIGEPS) r = rank(rind);

    if(moduloprint > 0 && cycles % moduloprint == 0) {
      srng_print(); 
      fprintf(stderr, "%d %g\n", cycles, neighborhood_size);
    }

    
    for(p = 0; p < pattern_length; p++) {
      
      pat = pattern[shuffle_next()];
      
      ptnopc = prototypes_no_class[(int)pat[pattern_dimension]];
      
      tmp = prototype_prepare_by_pattern(pat, 
          &corrects, &wrong,
           (neighborhood_size <= NEIGEPS || ptnopc == 1) ? 0 : 
            ((r < NEIGSPEEDTHRESH || ptnopc <= 5) ? 1 : 2));
      
      prototype_update(pat, tmp, corrects, wrong);

    }

    if(neighborhood_size > NEIGEPS) 
      neighborhood_size *= rate_neighborhood_annealing;
  }

  free(corrects);
}



int srng_error(void)
{

  static int calls;
  
  double *pat, d, min;

  int i, p, winner = 0 , correct, wrong,
      sum_correct[prototypes_no], sum_wrong[prototypes_no];

      
  if(++calls != moduloprint && moduloprint > 0) return 0;


  for(i = 0; i < prototypes_no; i++)
    sum_correct[i] = sum_wrong[i] = 0;
  
  
  for(p = 0; p < pattern_length; p++) {
  
    pat = pattern[p];
    
    min = DBL_MAX;

    for(i = 0; i < prototypes_no; i++) {
      
      d = prototype_patterndistance(prototypes[i], pat);
      
      if(d < min) {
        min = d;
        winner = i;
      }
    }
    
    if(prototypes[winner] -> cls == (int)pat[pattern_dimension]) 
      ++sum_correct[winner];
    else
      ++sum_wrong[winner];
    
  }
  
  correct = wrong = 0;
  
  for(i = prototypes_no - 1; i >= 0; i--) {
    printf("%d %d %g %d\n", i, sum_correct[i], 
                 sum_correct[i] / (double)(sum_correct[i] + sum_wrong[i]),
                 prototypes[i] -> cls);

    if(total_iterations < 0 && sum_correct[i] <= sum_wrong[i])
      prototypes[i] -> cls = -1;

    correct += sum_correct[i];
    wrong += sum_wrong[i];
  }

  printf("-1 %d %g %d\n\n", correct, correct / (double)(correct+wrong), calls);
  

  if(total_iterations < 0) { /* check out prototypes marked as bad (cls=-1) */

    for(i = 0; i < prototypes_no; i++) {
      
      if(prototypes[i] -> cls == -1) {
        
        prototype_free(prototypes[i]);

        for(p = i; p < prototypes_no-1; p++) 
          prototypes[p] = prototypes[p+1];

        --prototypes_no, --prototypes_lastindex;

        --i;
      }
    }
  }

  return 1;
}




/******************************************************************************/
/* no much checking done for main argument evaluation                         */
/******************************************************************************/


int main(int argc, char **argv)
{
 
  int print = 1;
  FILE *fp;
  
  total_iterations = (argc >= 3) ? atoi(argv[2]) : 1000;

  moduloprint = (argc >= 4) ? atoi(argv[3]) : 0;
  
  
  fp = (argv[1][0] == '-') ? stdin : fopen(argv[1], "rt");

  if(fp == NULL) panic("main(): cannot open initialization file");

  
  pattern_stdin();
  
  
  while(!feof(fp)) {
    
    srng_init(fp);
    
    if(total_iterations <= 0) 
      print = srng_error();
    else 
      srng_train(total_iterations);

    if(print) srng_print();

    ptfree();
  }

  fclose(fp);
  
  
  shuffle_index_free();
  pattern_free(pattern_length);
  rank(-1);                         /* free local table */

  return 0;
}
