# ztrans.awk
#
# Thu Feb 21 10:10:29  2002
# 
# z-transforms non-excluded data columns x:
#   calcs x'_i = (x_i-mean(x)) / stdev(x)
#
# -vIGNORE='RANGE'
# where RANGE is 
#  0. default "": none ignored
#  1. an index indicating the column to be ignored, 0 for the last column
#  2. a comma-separated list of indices
#  3. a range in a-b notation, meaning all indices between a and b including a,b
#  4. a comma-separated list of ranges
#  5. a comma-separated list of ranges and indices
#
# The filename must be given as argument.
# The standard input '-' is not allowed due to multi pass processing.
#
# Example:
#   awk -vIGNORE='1,4-6,10-12,0' -f ztrans.awk gurk.dat > gurkZ.dat 2> gurkZ_trans.dat
#
#   transforms all columns of gurk.dat excluding the last, 1,4,5,6,10,11,12;
#   saves the transformed data in gurkZ.dat, 
#   and the column parameters mean and standard deviation in gurkZ_trans.dat
#   (to be used in ztrans_inv.awk).

BEGIN {

  CONVFMT = "%.8g"

  STDERR = "cat >& 2"

  ARGC=4; ARGV[3] = ARGV[2] = ARGV[1]

  if(IGNORE != "") {
    
    split(IGNORE, rnges, ",")
    
    for(rng in rnges) {
      
      no = split(rnges[rng], fromto, "-")

      if(no == 1) 
        from = to = fromto[1] + 0.
      else if(no == 2)
        { from = fromto[1] + 0.; to = fromto[2] + 0.}
      else
        { printf "Error in IGNORE arg!" | STDERR; exit(0) }
      
      if(from > to) { h = from; from = to; to = h }

      for(i = from; i <= to; i++) ++ignore[i]
      
    }
  }
}


{ 
  if(!($0 ~ /(^#)|(^[ \t]*$)/)) {
  

    if(ARGIND == 1) {
    
      ++nr
    
      if(nf < NF) nf = NF

      for(i = 1; i <= nf; i++) 
        if(!(i in ignore || (i==nf && 0 in ignore)))
          mean[i] += $i
    }
    else if(ARGIND == 2) {

      if(firsttime == 0) {
      
        ++firsttime

        inv = 1. / nr
        
        for(i = 1; i <= nf; i++) 
          if(!(i in ignore || (i==nf && 0 in ignore)))
            mean[i] *= inv
      }

      for(i = 1; i <= nf; i++) 
        if(!(i in ignore || (i==nf && 0 in ignore)))
          stdev[i] += ($i-mean[i]) * ($i-mean[i])
    }
    else {

      for(i = 1; i <= nf; i++) {
      
        d = (stdev[i] <= 0.) ? 1 : sqrt(stdev[i] * inv)
        
        printf("%s ", (i in ignore || (i==nf && 0 in ignore)) ? ($i "") : 
                  ((($i-mean[i]) / d) ""))
      }
      printf("\n")
    }
  }
  else if(ARGIND == 1)
    print
}

END {

  for(i = 1; i <= nf; i++) 
    printf("%g ", mean[i] + 0.) | STDERR
  printf("\n") | STDERR
  
  for(i = 1; i <= nf; i++) 
    printf("%g ", stdev[i] * inv + 0.) | STDERR
  printf("\n") | STDERR
}
